@@ -1,4 +1,4 @@
-from __future__ import unicode_literals
+from __future__ import absolute_import, division, unicode_literals
 
 import getpass
 import logging
@@ -9,15 +9,15 @@ import re
 from rpaths import PosixPath, Path
 import scp
 import select
-import string
+import socket
 
-from tej.utils import string_types, iteritems, irange
+from tej.errors import InvalidDestination, QueueDoesntExist, \
+    QueueLinkBroken, QueueExists, JobAlreadyExists, JobNotFound, \
+    JobStillRunning, RemoteCommandFailure
+from tej.utils import unicode_, string_types, iteritems, irange, shell_escape
 
 
 __all__ = ['DEFAULT_TEJ_DIR',
-           'Error', 'InvalidDestionation', 'QueueDoesntExist',
-           'QueueLinkBroken', 'QueueExists', 'JobAlreadyExists', 'JobNotFound',
-           'JobStillRunning', 'RemoteCommandFailure',
            'parse_ssh_destination', 'destination_as_string',
            'ServerLogger', 'RemoteQueue']
 
@@ -25,80 +25,10 @@ __all__ = ['DEFAULT_TEJ_DIR',
 DEFAULT_TEJ_DIR = '~/.tej'
 
 
-class Error(Exception):
-    """Base class for exceptions.
-    """
-
-
-class InvalidDestionation(Error):
-    """Invalid SSH destination.
-    """
-    def __init__(self, msg="Invalid destination"):
-        super(InvalidDestionation, self).__init__(msg)
-
-
-class QueueDoesntExist(Error):
-    """Queue doesn't exist on the server.
-
-    `submit` and `setup` will create a queue on the server, but other commands
-    like `status` and `kill` expect it to be there.
-    """
-    def __init__(self, msg="Queue doesn't exist on the server"):
-        super(QueueDoesntExist, self).__init__(msg)
-
-
-class QueueLinkBroken(QueueDoesntExist):
-    """The chain of links is broken.
-
-    There is a link file on the server that doesn't point to anything.
-    """
-    def __init__(self, msg="Queue link chain is broken"):
-        super(QueueLinkBroken, self).__init__(msg)
-
-
-class QueueExists(Error):
-    """The queue whose creation was requested already exists.
-
-    A `force` argument (``--force``) is provided to replace it anyway.
-    """
-    def __init__(self, msg="Queue already exists"):
-        super(QueueExists, self).__init__(msg)
-
-
-class JobAlreadyExists(Error):
-    """A job with this name already exists on the server; submission failed.
-    """
-    def __init__(self, msg="Job already exists"):
-        super(JobAlreadyExists, self).__init__(msg)
-
-
-class JobNotFound(Error):
-    """A job with this name wasn't found on the server.
-    """
-    def __init__(self, msg="Job not found"):
-        super(JobNotFound, self).__init__(msg)
-
-
-class JobStillRunning(Error):
-    """An operation failed because the target job is still running.
-    """
-    def __init__(self, msg="Job is still running"):
-        super(JobStillRunning, self).__init__(msg)
-
-
-class RemoteCommandFailure(Exception):
-    """A failure that happened on the server.
-    """
-    def __init__(self, msg=None, command=None, ret=None):
-        if msg is None:
-            msg = "Command %r failed with status %d" % (command, ret)
-        super(RemoteCommandFailure, self).__init__(msg)
-
-
 logger = logging.getLogger('tej')
 
 
-def unique_names():
+def _unique_names():
     """Generates unique sequences of bytes.
     """
     characters = ("abcdefghijklmnopqrstuvwxyz"
@@ -108,31 +38,35 @@ def unique_names():
     while True:
         letters = [rng.choice(characters) for i in irange(10)]
         yield ''.join(letters)
-unique_names = unique_names()
+
+
+_unique_names = _unique_names()
 
 
 def make_unique_name():
     """Makes a unique (random) string.
     """
-    return next(unique_names)
+    return next(_unique_names)
 
 
-def shell_escape(s):
-    """Given bl"a, returns "bl\\"a".
+def escape_queue(s):
+    """Escapes the path to a queue, e.g. preserves ~ at the begining.
     """
-    if isinstance(s, bytes):
+    if isinstance(s, PosixPath):
+        s = unicode_(s)
+    elif isinstance(s, bytes):
         s = s.decode('utf-8')
-    if any(c in s for c in string.whitespace + '*$\\"\''):
-        return '"%s"' % (s.replace('\\', '\\\\')
-                          .replace('"', '\\"')
-                          .replace('$', '\\$'))
+    if s.startswith('~/'):
+        return '~/' + shell_escape(s[2:])
     else:
-        return s
+        return shell_escape(s)
 
 
 _re_ssh = re.compile(r'^'
                      r'(?:ssh://)?'              # 'ssh://' prefix
-                     r'(?:([a-zA-Z0-9_.-]+)@)?'  # 'user@'
+                     r'(?:([a-zA-Z0-9_.-]+)'     # 'user@'
+                     r'(?::([^ @]+))?'           # ':password'
+                     r'@)?'                      # '@'
                      r'([a-zA-Z0-9_.-]+)'        # 'host'
                      r'(?::([0-9]+))?'           # ':port'
                      r'$')
@@ -143,13 +77,15 @@ def parse_ssh_destination(destination):
     """
     match = _re_ssh.match(destination)
     if not match:
-        raise InvalidDestionation("Invalid destination: %s" % destination)
-    user, host, port = match.groups()
+        raise InvalidDestination("Invalid destination: %s" % destination)
+    user, password, host, port = match.groups()
     info = {}
     if user:
         info['username'] = user
     else:
         info['username'] = getpass.getuser()
+    if password:
+        info['password'] = password
     if port:
         info['port'] = int(port)
     info['hostname'] = host
@@ -158,12 +94,17 @@ def parse_ssh_destination(destination):
 
 
 def destination_as_string(destination):
+    if 'password' in destination:
+        user = '%s:%s' % (destination['username'], destination['password'])
+    else:
+        user = destination['username']
+
     if destination.get('port', 22) != 22:
-        return 'ssh://%s@%s:%d' % (destination['username'],
+        return 'ssh://%s@%s:%d' % (user,
                                    destination['hostname'],
                                    destination['port'])
     else:
-        return 'ssh://%s@%s' % (destination['username'],
+        return 'ssh://%s@%s' % (user,
                                 destination['hostname'])
 
 
@@ -187,27 +128,62 @@ class ServerLogger(object):
             self.data = []
 
     def message(self, data):
-        self.logger.warning(data)
+        self.logger.info(data)
+
+
+JOB_ID_CHARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ" \
+               "abcdefghijklmnopqrstuvwxyz" \
+               "0123456789_-+=@%:.,"
+
+
+def check_jobid(job_id):
+    if not all(c in JOB_ID_CHARS for c in job_id):
+        raise ValueError("Invalid job identifier")
 
 
 class RemoteQueue(object):
-    JOB_DONE = 0
-    JOB_RUNNING = 2
+    JOB_DONE = 'finished'
+    JOB_RUNNING = 'running'
+    JOB_INCOMPLETE = 'incomplete'
+    JOB_CREATED = 'created'
 
-    def __init__(self, destination, queue):
+    PROTOCOL_VERSION = 0, 2
+
+    def __init__(self, destination, queue,
+                 setup_runtime=None, need_runtime=None):
+        """Creates a queue object, that represents a job queue on a server.
+
+        :param destination: The address of the server, used to SSH into it.
+        :param queue: The pathname of the queue on the remote server. Something
+        like "~/.tej" is usually adequate. This will contain both the job info
+        and files, and the scripts used to manage it on the server side.
+        :param setup_runtime: The name of the runtime to deploy on the server
+        if the queue doesn't already exist. If None (default), it will
+        auto-detect what is appropriate (currently, `pbs` if the ``qsub``
+        command is available), and fallback on `default`. If `need_runtime` is
+        set, this should be one of the accepted values.
+        :param need_runtime: A list of runtime names that are acceptable. If
+        the queue already exists on the server and this argument is not None,
+        the installed runtime will be matched against it, and a failure will be
+        reported if it is not one of the provided values.
+        """
         if isinstance(destination, string_types):
-            try:
-                self.destination = parse_ssh_destination(destination)
-            except ValueError as e:
-                raise InvalidDestionation("Can't parse SSH destination %s" %
-                                          destination)
+            self.destination = parse_ssh_destination(destination)
         else:
             if 'hostname' not in destination:
-                raise InvalidDestionation("destination dictionary is missing "
-                                          "hostname")
+                raise InvalidDestination("destination dictionary is missing "
+                                         "hostname")
             self.destination = destination
+        if setup_runtime not in (None, 'default', 'pbs'):
+            raise ValueError("Selected runtime %r is unknown" % setup_runtime)
+        self.setup_runtime = setup_runtime
+        if need_runtime is not None:
+            self.need_runtime = set(need_runtime)
+        else:
+            self.need_runtime = None
         self.queue = PosixPath(queue)
-        self.ssh = None
+        self._queue = None
+        self._ssh = None
         self._connect()
 
     def server_logger(self):
@@ -222,17 +198,47 @@ class RemoteQueue(object):
     def destination_string(self):
         return destination_as_string(self.destination)
 
+    def _ssh_client(self):
+        """Gets an SSH client to connect with.
+        """
+        ssh = paramiko.SSHClient()
+        ssh.load_system_host_keys()
+        ssh.set_missing_host_key_policy(paramiko.RejectPolicy())
+        return ssh
+
     def _connect(self):
         """Connects via SSH.
         """
-        self.ssh = paramiko.SSHClient()
-        self.ssh.load_system_host_keys()
-        self.ssh.set_missing_host_key_policy(paramiko.RejectPolicy())
+        ssh = self._ssh_client()
         logger.debug("Connecting with %s",
-                     ', '.join('%s=%r' % i
-                               for i in iteritems(self.destination)))
-        self.ssh.connect(**self.destination)
+                     ', '.join('%s=%r' % (k, v if k != "password" else "***")
+                               for k, v in iteritems(self.destination)))
+        ssh.connect(**self.destination)
         logger.debug("Connected to %s", self.destination['hostname'])
+        self._ssh = ssh
+
+    def get_client(self):
+        """Gets the SSH client.
+
+        This will check that the connection is still alive first, and reconnect
+        if necessary.
+        """
+        if self._ssh is None:
+            self._connect()
+            return self._ssh
+        else:
+            try:
+                chan = self._ssh.get_transport().open_session()
+            except (socket.error, paramiko.SSHException):
+                logger.warning("Lost connection, reconnecting...")
+                self._ssh.close()
+                self._connect()
+            else:
+                chan.close()
+            return self._ssh
+
+    def get_scp_client(self):
+        return scp.SCPClient(self.get_client().get_transport())
 
     def _call(self, cmd, get_output):
         """Calls a command through the SSH connection.
@@ -242,23 +248,28 @@ class RemoteQueue(object):
         """
         server_err = self.server_logger()
 
-        chan = self.ssh.get_transport().open_session()
+        chan = self.get_client().get_transport().open_session()
         try:
             logger.debug("Invoking %r%s",
                          cmd, " (stdout)" if get_output else "")
             chan.exec_command('/bin/sh -c %s' % shell_escape(cmd))
             output = b''
-            while not chan.exit_status_ready():
+            while True:
                 r, w, e = select.select([chan], [], [])
-                if chan in r:
-                    if chan.recv_stderr_ready():
-                        data = chan.recv_stderr(1024)
-                        if data:
-                            server_err.append(data)
-                    if chan.recv_ready():
-                        data = chan.recv(1024)
-                        if get_output:
-                            output += data
+                if chan not in r:
+                    continue  # pragma: no cover
+                recvd = False
+                while chan.recv_stderr_ready():
+                    data = chan.recv_stderr(1024)
+                    server_err.append(data)
+                    recvd = True
+                while chan.recv_ready():
+                    data = chan.recv(1024)
+                    if get_output:
+                        output += data
+                    recvd = True
+                if not recvd and chan.exit_status_ready():
+                    break
             output = output.rstrip(b'\r\n')
             return chan.recv_exit_status(), output
         finally:
@@ -269,19 +280,19 @@ class RemoteQueue(object):
         """Calls a command through SSH.
         """
         ret, _ = self._call(cmd, False)
-        if ret != 0:
+        if ret != 0:  # pragma: no cover
             raise RemoteCommandFailure(command=cmd, ret=ret)
 
     def check_output(self, cmd):
         """Calls a command through SSH and returns its output.
         """
         ret, output = self._call(cmd, True)
-        if ret != 0:
+        if ret != 0:  # pragma: no cover
             raise RemoteCommandFailure(command=cmd, ret=ret)
         logger.debug("Output: %r", output)
         return output
 
-    def _resolve_queue(self, queue, depth=0):
+    def _resolve_queue(self, queue, depth=0, links=None):
         """Finds the location of tej's queue directory on the server.
 
         The `queue` set when constructing this `RemoteQueue` might be relative
@@ -293,46 +304,69 @@ class RemoteQueue(object):
         if depth == 0:
             logger.debug("resolve_queue(%s)", queue)
         answer = self.check_output(
-                'if [ -d %(queue)s ]; then '
-                '    cd %(queue)s; echo "dir: $(pwd)"; '
-                'elif [ -f %(queue)s ]; then '
-                '    cat %(queue)s; '
-                'else '
-                '    echo no; '
-                'fi' % {
-                    'queue': shell_escape(bytes(queue))})
+            'if [ -d %(queue)s ]; then '
+            '    cd %(queue)s; echo "dir"; cat version; pwd; '
+            'elif [ -f %(queue)s ]; then '
+            '    cat %(queue)s; '
+            'else '
+            '    echo no; '
+            'fi' % {
+                'queue': escape_queue(queue)})
         if answer == b'no':
             if depth > 0:
                 logger.debug("Broken link at depth=%d", depth)
             else:
                 logger.debug("Path doesn't exist")
             return None, depth
-        elif answer.startswith(b'dir: '):
-            new = PosixPath(answer[5:])
-            logger.debug("Found directory at %s, depth=%d", new, depth)
-            return new, depth
+        elif answer.startswith(b'dir\n'):
+            version, runtime, path = answer[4:].split(b'\n', 2)
+            try:
+                version = tuple(int(e)
+                                for e in version.decode('ascii', 'ignore')
+                                                .split('.'))
+            except ValueError:
+                version = 0, 0
+            if version[:2] != self.PROTOCOL_VERSION:
+                raise QueueExists(
+                    msg="Queue exists and is using incompatible protocol "
+                        "version %s" % '.'.join('%s' % e for e in version))
+            path = PosixPath(path)
+            runtime = runtime.decode('ascii', 'replace')
+            if self.need_runtime is not None:
+                if (self.need_runtime is not None and
+                        runtime not in self.need_runtime):
+                    raise QueueExists(
+                        msg="Queue exists and is using explicitely disallowed "
+                            "runtime %s" % runtime)
+            logger.debug("Found directory at %s, depth=%d, runtime=%s",
+                         path, depth, runtime)
+            return path, depth
         elif answer.startswith(b'tejdir: '):
             new = queue.parent / answer[8:]
             logger.debug("Found link to %s, recursing", new)
+            if links is not None:
+                links.append(queue)
             return self._resolve_queue(new, depth + 1)
-        logger.debug("Server returned %r", answer)
-        raise RemoteCommandFailure(msg="Remote command failed in unexpected "
-                                       "way")
+        else:  # pragma: no cover
+            logger.debug("Server returned %r", answer)
+            raise RemoteCommandFailure(msg="Queue resolution command failed "
+                                           "in unexpected way")
 
     def _get_queue(self):
         """Gets the actual location of the queue, or None.
         """
-        queue, depth = self._resolve_queue(self.queue)
-        if queue is None and depth > 0:
-            raise QueueLinkBroken
-        logger.debug("get_queue = %s", queue)
-        return queue
+        if self._queue is None:
+            self._links = []
+            queue, depth = self._resolve_queue(self.queue, links=self._links)
+            if queue is None and depth > 0:
+                raise QueueLinkBroken
+            self._queue = queue
+        return self._queue
 
     def setup(self, links=None, force=False, only_links=False):
         """Installs the runtime at the target location.
 
-        This will not replace an existing installation, unless it is a broken
-        chain of links or `force` is True.
+        This will not replace an existing installation, unless `force` is True.
 
         After installation, creates links to this installation at the specified
         locations.
@@ -344,8 +378,8 @@ class RemoteQueue(object):
             logger.info("Only creating links")
             for link in links:
                 self.check_call('echo "tejdir:" %(queue)s > %(link)s' % {
-                                'queue': shell_escape(str(self.queue)),
-                                'link': shell_escape(link)})
+                                'queue': escape_queue(self.queue),
+                                'link': escape_queue(link)})
             return
 
         queue, depth = self._resolve_queue(self.queue)
@@ -357,7 +391,7 @@ class RemoteQueue(object):
                     logger.info("Replacing link to %s...", queue)
                 else:
                     logger.info("Replacing existing queue...")
-                self.check_call('rm -Rf %s' % shell_escape(str(self.queue)))
+                self.check_call('rm -Rf %s' % escape_queue(self.queue))
             else:
                 if queue is not None and depth > 0:
                     raise QueueExists("Queue already exists (links to %s)\n"
@@ -373,29 +407,58 @@ class RemoteQueue(object):
 
         for link in links:
             self.check_call('echo "tejdir:" %(queue)s > %(link)s' % {
-                    'queue': shell_escape(str(queue)),
-                    'link': shell_escape(link)})
+                'queue': escape_queue(queue),
+                'link': escape_queue(link)})
 
     def _setup(self):
         """Actually installs the runtime.
         """
-        logger.info("Installing runtime at %s", self.queue)
-
         # Expands ~user in queue
-        output = self.check_output('echo %s' % shell_escape(str(self.queue)))
-        queue = PosixPath(output.rstrip(b'\r\n'))
-        logger.debug("Resolved to %s", queue)
+        if self.queue.path[0:1] == b'/':
+            queue = self.queue
+        else:
+            if self.queue.path[0:1] == b'~':
+                output = self.check_output('echo %s' %
+                                           escape_queue(self.queue))
+                queue = PosixPath(output.rstrip(b'\r\n'))
+            else:
+                output = self.check_output('pwd')
+                queue = PosixPath(output.rstrip(b'\r\n')) / self.queue
+            logger.debug("Resolved to %s", queue)
+
+        # Select runtime
+        if not self.setup_runtime:
+            # Autoselect
+            if self._call('which qsub', False)[0] == 0:
+                logger.debug("qsub is available, using runtime 'pbs'")
+                runtime = 'pbs'
+            else:
+                logger.debug("qsub not found, using runtime 'default'")
+                runtime = 'default'
+        else:
+            runtime = self.setup_runtime
+
+        if self.need_runtime is not None and runtime not in self.need_runtime:
+            raise ValueError("About to setup runtime %s but that wouldn't "
+                             "match explicitely allowed runtimes" % runtime)
+
+        logger.info("Installing runtime %s%s at %s",
+                    runtime,
+                    "" if self.setup_runtime else " (auto)",
+                    self.queue)
 
         # Uploads runtime
-        scp_client = scp.SCPClient(self.ssh.get_transport())
-        filename = pkg_resources.resource_filename('tej', 'remotes/default')
+        scp_client = self.get_scp_client()
+        filename = pkg_resources.resource_filename('tej',
+                                                   'remotes/%s' % runtime)
         scp_client.put(filename, str(queue), recursive=True)
         logger.debug("Files uploaded")
 
         # Runs post-setup script
-        self.check_call('/bin/sh %s' % (queue / 'commands' / 'setup'))
+        self.check_call('/bin/sh %s' % shell_escape(queue / 'commands/setup'))
         logger.debug("Post-setup script done")
 
+        self._queue = queue
         return queue
 
     def submit(self, job_id, directory, script=None):
@@ -404,21 +467,23 @@ class RemoteQueue(object):
         If the runtime is not there, it will be installed. If it is a broken
         chain of links, error.
         """
-        queue = self._get_queue()
-        if queue is None:
-            queue = self._setup()
-
         if job_id is None:
             job_id = '%s_%s_%s' % (Path(directory).unicodename,
                                    self.destination['username'],
                                    make_unique_name())
+        else:
+            check_jobid(job_id)
+
+        queue = self._get_queue()
+        if queue is None:
+            queue = self._setup()
 
         if script is None:
             script = 'start.sh'
 
         # Create directory
         ret, target = self._call('%s %s' % (
-                                 queue / 'commands' / 'new_job',
+                                 shell_escape(queue / 'commands/new_job'),
                                  job_id),
                                  True)
         if ret == 4:
@@ -429,31 +494,43 @@ class RemoteQueue(object):
         logger.debug("Server created directory %s", target)
 
         # Upload to directory
-        scp_client = scp.SCPClient(self.ssh.get_transport())
-        scp_client.put([str(p) for p in Path(directory).listdir()],
-                       str(target),
-                       recursive=True)
+        try:
+            scp_client = self.get_scp_client()
+            scp_client.put(str(Path(directory)),
+                           str(target),
+                           recursive=True)
+        except BaseException as e:
+            try:
+                self.delete(job_id)
+            except:
+                raise e
+            raise
         logger.debug("Files uploaded")
 
         # Submit job
-        self.check_call('%s %s %s %s' % (queue / 'commands' / 'submit',
-                                         job_id, target,
-                                         script))
+        self.check_call('%s %s %s %s' % (
+                        shell_escape(queue / 'commands/submit'),
+                        job_id, shell_escape(target),
+                        shell_escape(script)))
         logger.info("Submitted job %s", job_id)
         return job_id
 
     def status(self, job_id):
         """Gets the status of a previously-submitted job.
         """
+        check_jobid(job_id)
+
         queue = self._get_queue()
         if queue is None:
             raise QueueDoesntExist
 
-        ret, output = self._call('%s %s' % (queue / 'commands' / 'status',
-                                            job_id),
+        ret, output = self._call('%s %s' % (
+                                 shell_escape(queue / 'commands/status'),
+                                 job_id),
                                  True)
         if ret == 0:
             directory, result = output.splitlines()
+            result = result.decode('utf-8')
             return RemoteQueue.JOB_DONE, PosixPath(directory), result
         elif ret == 2:
             directory = output.splitlines()[0]
@@ -467,6 +544,8 @@ class RemoteQueue(object):
     def download(self, job_id, files, **kwargs):
         """Downloads files from server.
         """
+        check_jobid(job_id)
+
         if not files:
             return
         if isinstance(files, string_types):
@@ -490,7 +569,7 @@ class RemoteQueue(object):
         # Might raise JobNotFound
         status, target, result = self.status(job_id)
 
-        scp_client = scp.SCPClient(self.ssh.get_transport())
+        scp_client = self.get_scp_client()
         for filename in files:
             logger.info("Downloading %s", target / filename)
             if directory:
@@ -505,12 +584,15 @@ class RemoteQueue(object):
     def kill(self, job_id):
         """Kills a job on the server.
         """
+        check_jobid(job_id)
+
         queue = self._get_queue()
         if queue is None:
             raise QueueDoesntExist
 
-        ret, output = self._call('%s %s' % (queue / 'commands' / 'kill',
-                                            job_id),
+        ret, output = self._call('%s %s' % (
+                                 shell_escape(queue / 'commands/kill'),
+                                 job_id),
                                  False)
         if ret == 3:
             raise JobNotFound
@@ -521,12 +603,15 @@ class RemoteQueue(object):
     def delete(self, job_id):
         """Deletes a job from the server.
         """
+        check_jobid(job_id)
+
         queue = self._get_queue()
         if queue is None:
             raise QueueDoesntExist
 
-        ret, output = self._call('%s %s' % (queue / 'commands' / 'delete',
-                                            job_id),
+        ret, output = self._call('%s %s' % (
+                                 shell_escape(queue / 'commands/delete'),
+                                 job_id),
                                  False)
         if ret == 3:
             raise JobNotFound
@@ -543,7 +628,8 @@ class RemoteQueue(object):
         if queue is None:
             raise QueueDoesntExist
 
-        output = self.check_output('%s' % (queue / 'commands' / 'list'))
+        output = self.check_output('%s' %
+                                   shell_escape(queue / 'commands/list'))
 
         job_id, info = None, None
         for line in output.splitlines():
@@ -556,3 +642,29 @@ class RemoteQueue(object):
                     yield job_id, info
                 job_id = line
                 info = {}
+        if job_id is not None:
+            yield job_id, info
+
+    def cleanup(self, kill=False):
+        queue = self._get_queue()
+
+        if queue is not None:
+            # Kill jobs
+            for job_id, info in self.list():
+                if info['status'] == 'running':
+                    if not kill:
+                        raise JobStillRunning("Can't cleanup, some jobs are "
+                                              "still running")
+                    else:
+                        logger.info("Killing running job %s", job_id)
+                        self.kill(job_id)
+
+            # Remove queue
+            logger.info("Removing queue at %s", queue)
+            self.check_call('rm -rf -- %s' % shell_escape(queue))
+
+        # Remove links
+        for link in self._links:
+            self.check_call('rm -rf -- %s' % shell_escape(link))
+
+        return True
